# original code from https://github.com/SwinTransformer/Video-Swin-Transformer
# modified by Zilliz.

from functools import reduce
from operator import mul


def window_partition(x, window_size):
    """
    Args:
        x (`torch.Tensor`):
            Tensor with size of (b, d, h, w, c). b is batch size. d is time dimension size. h and w is frame size.
            c is channel size.
        window_size (`tuple[int]`):
            3d window size
    Returns:
        windows (`torch.Tensor`):
            Window partitioned tensor with size (B*num_windows, window_size*window_size, C)
    """
    b, d, h, w, c = x.shape
    x = x.view(b, d // window_size[0], window_size[0],
               h // window_size[1], window_size[1], w // window_size[2], window_size[2], c)
    windows = x.permute(0, 1, 3, 5, 2, 4, 6, 7).contiguous().view(-1, reduce(mul, window_size), c)
    return windows



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
import threading

from towhee.operator import Operator, SharedType
from towhee.engine.operator_loader import OperatorLoader


class _OperatorStorage:
    """
    Impl operator get and put by different shared_type.
    """
    def __init__(self):
        self._shared_type = None
        self._ops = []

    def op_available(self) -> bool:
        return self._shared_type is not None and len(self._ops) > 0

    def get(self):
        assert self._shared_type is not None and len(self._ops) != 0
        op = self._ops[-1]
        if self._shared_type != SharedType.Shareable:
            self._ops.pop()
        return op

    def put(self, op: Operator, force_put: bool = False):
        if self._shared_type is None:
            self._shared_type = op.shared_type

        if force_put or self._shared_type == SharedType.NotShareable:
            self._ops.append(op)

    def __len__(self):
        return len(self._ops)


class OperatorPool:
    """`OperatorPool` manages `Operator` creation, acquisition, release, and garbage
    collection. Each `TaskExecutor` has one `OperatorPool`.
    """
    def __init__(self, cache_path: str = None):
        self._op_loader = OperatorLoader(cache_path)
        self._all_ops = {}
        self._lock = threading.Lock()

    @staticmethod
    def _operator_id(hub_op_id: str, op_args: Dict[str, any], tag: str):
        if op_args:
            args_tup = tuple(sorted(op_args.items()))
        else:
            args_tup = ()
        return (hub_op_id, tag) + args_tup

    def __len__(self):
        num = 0
        for _, op_storage in self._all_ops.items():
            num += len(op_storage)
        return num

    def clear(self):
        self._all_ops = {}

    def acquire_op(self, hub_op_id: str, op_args: Dict[str, any], tag: str) -> Operator:
        """
        Instruct the `OperatorPool` to reserve and return the
        specified operator for use in the executor.

        Args:
            hub_op_id: (`str`)
            op_args: (`Dict[str, any]`)
                operator init parameters

        Returns:
            (`towhee.operator.Operator`)
                The operator instance reserved for the caller.
        """

        # Load the operator if the computed key does not exist in the operator
        # dictionary.
        op_key = OperatorPool._operator_id(hub_op_id, op_args, tag)
        with self._lock:
            storage = self._all_ops.get(op_key, None)
            if storage is None:
                storage = _OperatorStorage()
                self._all_ops[op_key] = storage

            if not storage.op_available():
                op = self._op_loader.load_operator(hub_op_id, op_args, tag)
                op.key = op_key
                storage.put(op, True)
            return storage.get()

    def release_op(self, op: Operator):
        """Releases the specified operator and all associated resources back to the
        `OperatorPool`.

        Args:
            op: (`towhee.Operator`)
                `Operator` instance to add back into the operator pool.
        """
        with self._lock:
            storage = self._all_ops[op.key]
            storage.put(op)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy
from typing import Tuple

from towhee._types import Image
from towhee.types import AudioFrame
from towhee.hparam import param_scope
from towhee.functional.entity import Entity

# pylint: disable=dangerous-default-value


def get_df_on_columns(self, index: Tuple[str]):
    # pylint: disable=import-outside-toplevel
    from towhee.utils.pandas_utils import pandas as pd

    def inner(entity: Entity):
        data = {}
        for feature in index:
            data[feature] = getattr(entity, feature)
        return data

    data = map(inner, self)
    df = pd.DataFrame(data)
    return df


def calc_df(df, feature: str, target: str):
    # pylint: disable=import-outside-toplevel
    from towhee.utils.pandas_utils import pandas as pd
    lst = []
    df[feature] = df[feature].fillna('NULL')

    if target:
        for i in range(df[feature].nunique()):
            val = list(df[feature].unique())[i]
            lst.append([feature,                                                         # Variable
                        val,                                                             # Value
                        df[df[feature] == val].count()[feature],                         # All
                        df[(df[feature] == val) & (df[target] == 0)].count()[feature],   # Good (think: Fraud == 0)
                        df[(df[feature] == val) & (df[target] == 1)].count()[feature]])  # Bad (think: Fraud == 1)

        data = pd.DataFrame(lst, columns=['Variable', 'Value', 'All', 'Good', 'Bad'])
        data['Share'] = data['All'] / data['All'].sum()
        data['Bad Rate'] = data['Bad'] / data['All']
        data['Distribution Good'] = (data['All'] - data['Bad']) / (data['All'].sum() - data['Bad'].sum())
        data['Distribution Bad'] = data['Bad'] / data['Bad'].sum()
        data['WoE'] = numpy.log(data['Distribution Good'] / data['Distribution Bad'])

        data = data.replace({'WoE': {numpy.inf: 0, -numpy.inf: 0}})
        data['IV'] = data['WoE'] * (data['Distribution Good'] - data['Distribution Bad'])

        data = data.sort_values(by=['Variable', 'Value'], ascending=[True, True])
        data.index = range(len(data.index))

        iv = data['IV'].sum()
        print(f'Variable: {feature}\'s IV sum is: {iv}')
        return data
    else:
        for i in range(df[feature].nunique()):
            val = list(df[feature].unique())[i]
            lst.append([feature,                                                         # Variable
                        val,                                                             # Value
                        df[df[feature] == val].count()[feature]])                        # All
        data = pd.DataFrame(lst, columns=['Variable', 'Value', 'All'])
        data['Share'] = data['All'] / data['All'].sum()
        return data


def _feature_summarize_callback(self):
    # pylint: disable=import-outside-toplevel
    from towhee.utils.pandas_utils import pandas as pd

    def wrapper(_: str, index, *arg, **kws):
        if isinstance(index, str):
            index = (index,)
        index = list(index)
        if arg:
            kws['target'], = arg

        target = None
        if 'target' in kws:
            target = kws.pop('target')
            index.append(target)

        df = get_df_on_columns(self, index)
        summarize = pd.DataFrame()
        if target:
            index.remove(target)
        for feature in index:
            data = calc_df(df, feature, target)
            summarize = summarize.append(data)

        # pylint: disable=import-outside-toplevel
        from towhee.utils import ipython_utils
        ipython_utils.display(summarize)

    return wrapper


def _plot_callback(self):
    # pylint: disable=unused-argument
    def wrapper(_: str, index, *arg, **kws):
        if isinstance(index, str):
            index = (index,)
        df = get_df_on_columns(self, index)
        if 'kind' not in kws:
            kws.update(kind='hist')
        df.plot(**kws)

    return wrapper


class DisplayMixin:
    """
    Mixin for display data.
    """

    def __init__(self):
        super().__init__()
        self.feature_summarize = param_scope().callholder(_feature_summarize_callback(self))
        self.plot = param_scope().callholder(_plot_callback(self))

    def as_str(self):
        return self._factory(map(str, self._iterable))

    def show(self, limit=5, header=None, tablefmt='html', formatter={}):
        """
        Print the first n lines of a DataCollection.

        Args:
            limit (`int`):
                The number of lines to print. Default value is 5.
                Print all if limit is non-positive.
            header (list of `str`):
                The field names.
            tablefmt (`str`):
                The format of the output, support html, plain, grid.
        """

        contents = [x for i, x in enumerate(self._iterable) if i < limit]

        if all(isinstance(x, Entity) for x in contents):
            header = tuple(contents[0].to_dict()) if not header else header
            data = [list(x.to_dict().values()) for x in contents]
        else:
            data = [[x] for x in contents]

        table_display(to_printable_table(data, header, tablefmt, formatter))


def table_display(table, tablefmt='html'):
    """
    Display table

    Args:
        table (`str`):
            Table in printable format, such as HTML.
        tablefmt (`str`):
            The format of the output, support html, plain, grid.
    """
    # pylint: disable=import-outside-toplevel
    from towhee.utils import ipython_utils

    if tablefmt == 'html':
        ipython_utils.display(ipython_utils.HTML(table))
    elif tablefmt in ('plain', 'grid'):
        print(table)
    else:
        raise ValueError('unsupported table format %s' % tablefmt)


def to_printable_table(data, header=None, tablefmt='html', formatter={}):
    """
    Convert two dimensional data structure into printable table

    Args:
        data (`List[List]`, or `List[Dict]`):
            The data filled into table. If a list of dict is given, keys are used as column names.
        header (`Iterable[str]`):
            The name of columns defined by users.
        tablefmt (`str`):
            The format of the output, support html, plain, grid.
    """
    header = [] if not header else header

    if tablefmt == 'html':
        return to_html_table(data, header, formatter)
    elif tablefmt in ('plain', 'grid'):
        return to_plain_table(data, header, tablefmt)

    raise ValueError('unsupported table format %s' % tablefmt)


def to_plain_table(data, header, tablefmt):
    """
    Convert two dimensional data structure into plain table

    Args:
        data (`List[List]`, or `List[Dict]`):
            The data filled into table. If a list of dict is given, keys are used as column names.
        header (`Iterable[str]`):
            The name of columns defined by users.
        tablefmt (`str`):
            The format of the output, support plain, grid.
    """
    # pylint: disable=import-outside-toplevel
    from tabulate import tabulate

    tb_contents = [[_to_plain_cell(x) for x in r] for r in data]
    return tabulate(tb_contents, headers=header, tablefmt=tablefmt)


def _to_plain_cell(data):
    if isinstance(data, str):
        return _text_brief(data)
    if isinstance(data, Image):
        return _image_brief(data)
    elif isinstance(data, AudioFrame):
        return _audio_frame_brief(data)
    elif isinstance(data, numpy.ndarray):
        return _ndarray_brief(data)

    elif isinstance(data, (list, tuple)):
        if all(isinstance(x, str) for x in data):
            return _list_brief(data, _text_brief)
        elif all(isinstance(x, Image) for x in data):
            return _list_brief(data, _image_brief)
        elif all(isinstance(x, AudioFrame) for x in data):
            return _list_brief(data, _audio_frame_brief)
        elif all(isinstance(x, numpy.ndarray) for x in data):
            return _list_brief(data, _ndarray_brief)
    return _default_brief(data)


def to_html_table(data, header, formatter={}):
    """
    Convert two dimensional data structure into html table

    Args:
        data (`List[List]`, or `List[Dict]`):
            The data filled into table. If a list of dict is given, keys are used as column names.
        header (`Iterable[str]`):
            The name of columns defined by users.
    """
    tb_style = 'style="border-collapse: collapse;"'
    th_style = 'style="text-align: center; font-size: 130%; border: none;"'
    td_style = 'style="text-align: center; border-right: solid 1px #D3D3D3; border-left: solid 1px #D3D3D3;"'

    str_2_callback = {
        'text': _text_brief,
        'image': _image_to_html_cell
    }

    trs = []
    trs.append('<tr>' + ' '.join(['<th ' + th_style + '>' + x + '</th>' for x in header]) + '</tr>')

    to_html_callback = {}
    for i, field in enumerate(header):
        cb = formatter.get(field, None)
        if cb is None:
            cb = _to_html_cell
        elif isinstance(cb, str):
            cb = str_2_callback.get(cb, _to_html_cell)
        to_html_callback[i] = cb
    for r in data:
        trs.append(
            '<tr>' + ' '.join(['<td ' + td_style + '>' + to_html_callback.get(i, _to_html_cell)(x) + '</td>' for i, x in enumerate(r)]) + '</tr>'
        )
    return '<table ' + tb_style + '>' + ' '.join(trs) + '</table>'


def _to_html_cell(data):
    if isinstance(data, str):
        return _text_brief(data)
    if isinstance(data, Image):
        return _image_to_html_cell(data)
    elif isinstance(data, AudioFrame):
        return _audio_frame_brief(data)
    elif isinstance(data, numpy.ndarray):
        return _ndarray_brief(data)

    elif isinstance(data, (list, tuple)):
        if all(isinstance(x, str) for x in data):
            return _list_brief(data, _text_brief)
        elif all(isinstance(x, Image) for x in data):
            return _images_to_html_cell(data)
        elif all(isinstance(x, AudioFrame) for x in data):
            return _list_brief(data, _audio_frame_brief)
        elif all(isinstance(x, numpy.ndarray) for x in data):
            return _list_brief(data, _ndarray_brief)
    return _default_brief(data)


def _image_to_html_cell(img, width=128, height=128):
    # pylint: disable=import-outside-toplevel
    from towhee.utils.cv2_utils import cv2
    import base64

    _, img_encode = cv2.imencode('.JPEG', img)
    src = 'src="data:image/jpeg;base64,' + base64.b64encode(img_encode).decode() + '" '
    w = 'width = "' + str(width) + '" '
    h = 'height = "' + str(height) + '" '

    return '<img ' + src + w + h + '>'


def _images_to_html_cell(imgs, width=128, height=128):
    return ' '.join([_image_to_html_cell(x, width, height) for x in imgs])


def _ndarray_brief(array, maxlen=3):
    head_vals = [repr(v) for i, v in enumerate(array.flatten()) if i < maxlen]
    if len(array.flatten()) > maxlen:
        head_vals.append('...')
    shape = 'shape=' + repr(array.shape)

    return '[' + ', '.join(head_vals) + '] ' + shape


def _image_brief(img):
    return str(img)


def _audio_frame_brief(frame):
    return str(frame)


def _text_brief(text, maxlen=32):
    if len(text) > maxlen:
        return text[:maxlen] + '...'
    else:
        return text


def _list_brief(data, str_method, maxlen=4):
    head_vals = [str_method(x) for i, x in enumerate(data) if i < maxlen]
    if len(data) > maxlen:
        head_vals.append('...')
    return '[' + ','.join(head_vals) + ']' + ' len=' + str(len(data))


def _default_brief(data, maxlen=128):
    s = repr(data)
    return s[:maxlen] + '...' if len(s) > maxlen else s


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import List, Tuple, Union
from towhee.dataframe import DataFrame
from towhee.engine.engine import Engine, start_engine
from towhee.engine.pipeline import Pipeline
from towhee.pipeline_format import OutputFormat
from towhee.hub.file_manager import FileManagerConfig, FileManager
from towhee.engine import register
from towhee.engine.factory import ops
from towhee.hparam import param_scope
from towhee.hparam import HyperParameter as Document
from towhee.functional import DataCollection, State, Entity

from towhee.functional import glob, glob_zip
from towhee.functional import read_csv, read_json, read_camera
from towhee.functional import from_zip, from_df
from towhee.functional import dc

__all__ = [
    'DEFAULT_PIPELINES',
    'pipeline',
    'ops',
    'register',
    'param_scope',
    'Build',
    'Inject',
    'dataset',
    'Document',
    'DataCollection',
    'State',
    'Entity',
    #
    'glob',
    'glob_zip',
    'read_csv',
    'read_json',
    'read_camera',
    'from_zip',
    'from_df',
    'dc'
]

DEFAULT_PIPELINES = {
    'image-embedding': 'towhee/image-embedding-resnet50',
    'image-encoding': 'towhee/image-embedding-resnet50',  # TODO: add encoders
    'music-embedding': 'towhee/music-embedding-vggish',
    'music-encoding': 'towhee/music-embedding-clmr',  # TODO: clmr -> encoder
}

_PIPELINE_CACHE_ENV = 'PIPELINE_CACHE'


class _PipelineWrapper:
    """
    A wrapper class around `Pipeline`.

    The class prevents users from having to create `DataFrame` instances by hand.

    Args:
        pipeline (`towhee.Pipeline`):
            Base `Pipeline` instance for which this object will provide a wrapper for.
    """
    def __init__(self, pipeline_: Pipeline):
        self._pipeline = pipeline_

    def __call__(self, *args) -> List[Tuple]:
        """
        Wraps the input arguments around a `Dataframe` for Pipeline.__call__(). For
        example:
        ```
        >>> p = pipeline('some-pipeline')
        >>> result = p(arg0, arg1)
        ```
        """
        if not args:
            raise RuntimeError('Input data is empty')

        cols = []
        vargs = []
        for i, arg in enumerate(args):
            vtype = type(arg).__name__
            cols.append(('Col_' + str(i), str(vtype)))
            vargs.append(arg)
        vargs = tuple(vargs)

        # Process the data through the pipeline.
        in_df = DataFrame('_in_df', cols)
        in_df.put(vargs)
        out_df = self._pipeline(in_df)
        format_handler = OutputFormat.get_format_handler(self._pipeline.pipeline_type)
        return format_handler(out_df)

    def __repr__(self) -> str:
        return repr(self._pipeline)

    @property
    def pipeline(self) -> Pipeline:
        return self._pipeline


def pipeline(pipeline_src: str, tag: str = 'main', install_reqs: bool = True):
    """
    Entry method which takes either an input task or path to an operator YAML.

    A `Pipeline` object is created (based on said task) and subsequently added to the
    existing `Engine`.

    Args:
        pipeline_src (`str`):
            pipeline name or YAML file location to use.
        tag (`str`):
            Which tag to use for operators/pipelines on hub, defaults to `main`.
        install_reqs (`bool`):
            Whether to download the python packages if a requirements.txt file is included in the repo.

    Returns
        (`typing.Any`)
            The `Pipeline` output.
    """
    start_engine()

    if os.path.isfile(pipeline_src):
        yaml_path = pipeline_src
    else:
        fm = FileManager()
        p_repo = DEFAULT_PIPELINES.get(pipeline_src, pipeline_src)
        yaml_path = fm.get_pipeline(p_repo, tag, install_reqs)

    engine = Engine()
    pipeline_ = Pipeline(str(yaml_path))
    with param_scope() as hp:
        if not hp().towhee.dry_run(False):
            engine.add_pipeline(pipeline_)

    return _PipelineWrapper(pipeline_)


def dataset(name: str, *args, **kwargs) -> 'TorchDataSet':
    """
    Get a dataset by name, and pass into the custom params.

    Args:
        name (`str`):
            Name of a dataset.
        *args (`Any`):
            Arguments of the dataset construct method.
        **kwargs (`Any`):
            Keyword arguments of the dataset construct method.

    Returns:
        (`TorchDataSet`)
            The corresponding `TorchDataSet`.

    Examples:
        >>> from towhee import dataset
        >>> type(dataset('fake', size=10))
        <class 'towhee.data.dataset.dataset.TorchDataSet'>
    """
    from torchvision import datasets  # pylint: disable=import-outside-toplevel
    from towhee.data.dataset.dataset import TorchDataSet  # pylint: disable=import-outside-toplevel
    dataset_construct_map = {
        'mnist': datasets.MNIST, 'cifar10': datasets.cifar.CIFAR10, 'fake': datasets.FakeData
        # 'imdb': IMDB  # ,()
    }
    torch_dataset = dataset_construct_map[name](*args, **kwargs)
    return TorchDataSet(torch_dataset)


class Build:
    """
    Build a pipeline with template variables.

    A pipeline template is a yaml file contains `template variables`,
    which will be replaced by `variable values` when createing pipeline instance.

    Examples:
    ```yaml
    name: template_name
    variables:                           <<-- define variables and default values
        template_variable_1: default_value_1
        template_variable_2: default_value_2
    ....

    operator:
        function: {template_variable_1}  <<-- refer to the variable by name
    ```

    You can specialize template variable values with the following code:

    >>> pipe = Build(template_variable_1='new_value').pipeline('pipeline_name')
    """
    def __init__(self, **kws) -> None:
        self._kws = kws

    def pipeline(self, *arg, **kws):
        with param_scope() as hp:
            hp().variables = self._kws
            return pipeline(*arg, **kws)


class Inject:
    """
    Build a pipeline by operator injection.

    Injecting is another pipeline templating mechanism that allows the use to modify
    the pipeline directly without declaring template variables.

    Examples:
    ```yaml
    operators:
    - name: operator_1
      function: namespace/operator_1         <<-- injection point
      ...
    - name: operator_2
      function: namespace/operator_2
    ```

    You can modify the pipeline directly by:

    >>> pipe = Inject(operator_1 = dict(function='my_namespace/my_op')).pipeline('pipeline_name')

    and the value at the injection point is replace by the `Inject` API.
    """
    def __init__(self, **kws) -> None:
        self._injections = {}
        for k, v in kws.items():
            self._injections[k] = v

    def pipeline(self, *arg, **kws):
        with param_scope() as hp:
            hp().injections = self._injections
            return pipeline(*arg, **kws)


def _pipeline_callback(name, index, *arg, **kws):
    name = name.replace('.', '/').replace('_', '-')
    _ = index
    return Build(**kws).pipeline(name, *arg)


pipes = param_scope().callholder(_pipeline_callback)
"""
Entry point for creating pipeline instances, for example:

>>> pipe_instance = pipes.my_namespace.my_pipeline_name(template_variable_1=xxx, template_variable_2=xxx)

An instance of `my_namespace`/`my_pipeline_name` is created, and template variables in the pipeline,z
`template_variable_1` and  `template_variable_2` are replaced with given values.
"""


def plot(img1: Union[str, list], img2: list = None):
    from towhee.utils.plot_utils import plot_img  # pylint: disable=C
    if not img2:
        plot_img(img1)
    else:
        assert len(img1) == len(img2)
        for i in range(len(img1)):
            plot_img(img1[i])
            plot_img(img2[i])



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .data_collection import DataCollection
from .entity import Entity
from .option import Option, Some, Empty
from towhee.hparam import HyperParameter as State

from towhee.hparam.hyperparameter import CallTracer


class GlobImpl(CallTracer):
    """
    Return a DataCollection of paths matching a pathname pattern.
    Examples:

    1. create a simple data collection;

    >>> import towhee
    >>> towhee.glob('*.jpg').to_list()
    ['a.jpg', 'b.jpg]

    2. create a data collection of structural data.

    >>> towhee.glob['path']('*.jpg').to_list()
    [<Entity dict_keys(['path'])>, <Entity dict_keys(['path'])>]
    """

    def __init__(self, callback=None, path=None, index=None):
        super().__init__(callback=callback, path=path, index=index)


class GlobZipImpl(CallTracer):
    """
    Return a DataCollection of files matching a pathname pattern from a zip archive.
    Examples:

    1. create a simple data collection;

    >>> import towhee
    >>> towhee.glob_zip('somefile.zip', '*.jpg').to_list()
    ['a.jpg', 'b.jpg]

    2. create a data collection of structural data.

    >>> towhee.glob_zip['path']('somefile.zip', '*.jpg').to_list()
    [<Entity dict_keys(['path'])>, <Entity dict_keys(['path'])>]
    """

    def __init__(self, callback=None, path=None, index=None):
        super().__init__(callback=callback, path=path, index=index)


class DCImpl(CallTracer):
    """
    Return a DataCollection.

    Examples:

    1. create a simple data collection;

    >>> import towhee
    >>> towhee.dc([0, 1, 2]).to_list()
    [0, 1, 2]

    2. create a data collection of structural data.

    >>> towhee.dc['column']([0, 1, 2]).to_list()
    [<Entity dict_keys(['column'])>, <Entity dict_keys(['column'])>, <Entity dict_keys(['column'])>]
    """

    def __init__(self, callback=None, path=None, index=None):
        super().__init__(callback=callback, path=path, index=index)


def stream(iterable):
    return DataCollection.stream(iterable)


def unstream(iterable):
    return DataCollection.unstream(iterable)


read_csv = DataCollection.from_csv

read_json = DataCollection.from_json

read_camera = DataCollection.from_camera

from_zip = DataCollection.from_zip


def from_df(df):
    return DataCollection.stream(
        df.iterrows()).map(lambda x: Entity(**x[1].to_dict()))


def _glob_call_back(_, index, *arg, **kws):
    if index is not None:
        return DataCollection.from_glob(
            *arg, **kws).map(lambda x: Entity(**{index: x}))
    else:
        return DataCollection.from_glob(*arg, **kws)


def _glob_zip_call_back(_, index, *arg, **kws):
    if index is not None:
        return from_zip(*arg, **kws).map(lambda x: Entity(**{index: x}))
    else:
        return from_zip(*arg, **kws)


def _dc_call_back(_, index, *arg, **kws):
    if index is not None:
        return DataCollection.stream(*arg, **kws).map(lambda x: Entity(**{index: x}))
    else:
        return DataCollection.stream(*arg, **kws)


glob = GlobImpl(_glob_call_back)

glob_zip = GlobZipImpl(_glob_zip_call_back)

dc = DCImpl(_dc_call_back)



# Copyright 2021 Microsoft . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
import torch
from torch import nn
from towhee.models.layers.mlp import Mlp
from towhee.models.layers.droppath import DropPath
from towhee.models.utils.general_utils import to_2tuple
from .utils import window_reverse, window_partition
from .window_attention import WindowAttention


class SwinTransformerBlock(nn.Module):
    r""" Swin Transformer Block.
    Args:
        dim (int): Number of input channels.
        input_resolution (tuple[int]): Input resulotion.
        num_heads (int): Number of attention heads.
        window_size (int): Window size.
        shift_size (int): Shift size for SW-MSA.
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
        qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
        drop (float, optional): Dropout rate. Default: 0.0
        attn_drop (float, optional): Attention dropout rate. Default: 0.0
        drop_path (float, optional): Stochastic depth rate. Default: 0.0
        act_layer (nn.Module, optional): Activation layer. Default: nn.GELU
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(self, dim, input_resolution, num_heads, window_size=7, shift_size=0,
                 mlp_ratio=4., qkv_bias=True, drop=0., attn_drop=0., drop_path=0.,
                 act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.dim = dim
        self.input_resolution = input_resolution
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        self.mlp_ratio = mlp_ratio
        if min(self.input_resolution) <= self.window_size:
            # if window size is larger than input resolution, we don't partition windows
            self.shift_size = 0
            self.window_size = min(self.input_resolution)
        assert 0 <= self.shift_size < self.window_size, 'shift_size must in 0-window_size'

        self.norm1 = norm_layer(dim)
        self.attn = WindowAttention(
            dim, window_size=to_2tuple(self.window_size), num_heads=num_heads, qkv_bias=qkv_bias,
            attn_drop=attn_drop, proj_drop=drop)

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        if self.shift_size > 0:
            # calculate attention mask for SW-MSA
            h, w = self.input_resolution
            img_mask = torch.zeros((1, h, w, 1))  # 1 H W 1
            h_slices = (slice(0, -self.window_size),
                        slice(-self.window_size, -self.shift_size),
                        slice(-self.shift_size, None))
            w_slices = (slice(0, -self.window_size),
                        slice(-self.window_size, -self.shift_size),
                        slice(-self.shift_size, None))
            cnt = 0
            for h in h_slices:
                for w in w_slices:
                    img_mask[:, h, w, :] = cnt
                    cnt += 1

            mask_windows = window_partition(img_mask, self.window_size)  # nW, window_size, window_size, 1
            mask_windows = mask_windows.view(-1, self.window_size * self.window_size)
            attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
            attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(attn_mask == 0, float(0.0))
        else:
            attn_mask = None

        self.register_buffer('attn_mask', attn_mask)

    def forward(self, x):
        h, w = self.input_resolution
        b, l, c = x.shape
        assert l == h * w, 'input feature has wrong size'

        shortcut = x
        x = self.norm1(x)
        x = x.view(b, h, w, c)

        # cyclic shift
        if self.shift_size > 0:
            shifted_x = torch.roll(x, shifts=(-self.shift_size, -self.shift_size), dims=(1, 2))
        else:
            shifted_x = x

        # partition windows
        x_windows = window_partition(shifted_x, self.window_size)  # nW*B, window_size, window_size, C
        x_windows = x_windows.view(-1, self.window_size * self.window_size, c)  # nW*B, window_size*window_size, C

        # W-MSA/SW-MSA
        attn_windows = self.attn(x_windows, mask=self.attn_mask)  # nW*B, window_size*window_size, C

        # merge windows
        attn_windows = attn_windows.view(-1, self.window_size, self.window_size, c)
        shifted_x = window_reverse(attn_windows, self.window_size, h, w)  # B H' W' C

        # reverse cyclic shift
        if self.shift_size > 0:
            x = torch.roll(shifted_x, shifts=(self.shift_size, self.shift_size), dims=(1, 2))
        else:
            x = shifted_x
        x = x.view(b, h * w, c)

        # FFN
        x = shortcut + self.drop_path(x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union
from pathlib import Path

from towhee.functional.entity import Entity


class DatasetMixin:
    """
    Mixin for dealing with dataset
    """

    # pylint: disable=import-outside-toplevel
    @classmethod
    def from_glob(cls, pattern):
        """
        generate a file list with `pattern`
        """
        from glob import glob
        return cls.stream(glob(pattern))

    @classmethod
    def from_zip(cls, url, pattern, mode='r'):
        """load files from url/path.

        Args:
            zip_src (`Union[str, path]`):
                The path leads to the image.
            pattern (`str`):
                The filename pattern to extract.
            mode (str):
                file open mode.

        Returns:
            (File): The file handler for file in the zip file.
        """
        from towhee.utils.repo_normalize import RepoNormalize
        from io import BytesIO
        from zipfile import ZipFile
        from glob import fnmatch

        from urllib.request import urlopen

        def inner():
            if RepoNormalize(str(url)).url_valid():
                with urlopen(url) as zip_file:
                    zip_path = BytesIO(zip_file.read())
            else:
                zip_path = str(Path(url).resolve())
            with ZipFile(zip_path, 'r') as zfile:
                file_list = zfile.namelist()
                path_list = fnmatch.filter(file_list, pattern)
                for path in path_list:
                    with zfile.open(path, mode=mode) as f:
                        yield f.read()

        return cls.stream(inner())

    @classmethod
    def from_camera(cls, device_id=0, limit=-1):
        """
        read images from a camera.
        """
        from towhee.utils.cv2_utils import cv2
        cnt = limit

        def inner():
            nonlocal cnt
            cap = cv2.VideoCapture(device_id)
            while cnt != 0:
                retval, im = cap.read()
                if retval:
                    yield im
                    cnt -= 1

        return cls.stream(inner())

    @classmethod
    def from_json(cls, json_path: Union[str, Path], encoding: str = 'utf-8'):
        import json

        def inner():
            with open(json_path, 'r', encoding=encoding) as f:
                string = f.readline()
                while string:
                    data = json.loads(string)
                    string = f.readline()
                    yield Entity(**data)

        return cls.stream(inner())

    @classmethod
    def from_csv(cls, csv_path: Union[str, Path], encoding: str = 'utf-8-sig'):
        import csv

        def inner():
            with open(csv_path, 'r', encoding=encoding) as f:
                data = csv.DictReader(f)
                for line in data:
                    yield Entity(**line)

        return cls.stream(inner())

    def random_sample(self):
        # core API already exists
        pass

    def filter_data(self):
        # core API already exists
        pass

    # pylint: disable=dangerous-default-value
    def split_train_test(self, size: list = [0.9, 0.1], **kws):
        """
        Split DataCollection to train and test data.

        Args:
            size (`list`):
                The size of the train and test.

        Examples:

        >>> from towhee.functional import DataCollection
        >>> dc = DataCollection.range(10)
        >>> train, test = dc.split_train_test(shuffle=False)
        >>> train.to_list()
        [0, 1, 2, 3, 4, 5, 6, 7, 8]
        >>> test.to_list()
        [9]
        """
        from towhee.utils import sklearn_utils
        train_size = size[0]
        test_size = size[1]
        train, test = sklearn_utils.train_test_split(list(self), train_size=train_size, test_size=test_size, **kws)
        return self._factory(train), self._factory(test)

    def save_json(self):
        pass

    def save_image(self):
        pass

    def save_csv(self):
        pass

    def get_info(self):
        pass


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False)



# Copyright 2021 biubug6 . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.

#adapted from https://github.com/biubug6/Pytorch_Retinaface
#from collections import OrderedDict
from typing import Tuple, Dict

import numpy as np
import torch
from torch import nn
import torch.nn.functional as F
import torchvision
from torchvision import models
from torchvision.models import _utils

from towhee.models.retina_face.ssh import SSH
from towhee.models.retina_face.mobilenet_v1 import MobileNetV1
from towhee.models.retina_face.retinaface_fpn import RetinaFaceFPN
from towhee.models.retina_face.heads import ClassHead, BboxHead, LandmarkHead
from towhee.models.retina_face.prior_box import PriorBox
from towhee.models.retina_face.utils import decode, decode_landm

class RetinaFace(nn.Module):
    """
    ReitinaFace

    RetinaFace: Single-stage Dense Face Localisation in the Wild.
    Described in https://arxiv.org/abs/1905.00641.

    Args:
        cfg (`Dict`):
            Network related settings.
        phase (`str`):
            train or test.
    """
    def __init__(self, cfg: Dict=None, phase: str='train'):
        super().__init__()
        self.phase = phase
        backbone = None
        self.cfg = cfg
        if cfg['name'] == 'mobilenet0.25':
            backbone = MobileNetV1()
            #if cfg['pretrain']:
            #    checkpoint = torch.load('./pretrained_weights/mobilenetV1X0.25_pretrain.tar', map_location=torch.device('cpu'))
            #    new_state_dict = OrderedDict()
            #    for k, v in checkpoint['state_dict'].items():
            #        name = k[7:]  # remove module.
            #        new_state_dict[name] = v
            #    # load params
            #    backbone.load_state_dict(new_state_dict)
        elif cfg['name'] == 'Resnet50':
            #backbone = models.resnet50(pretrained=cfg['pretrain'])
            backbone = models.resnet50(False)

        self.body = _utils.IntermediateLayerGetter(backbone, cfg['return_layers'])
        in_channels_stage2 = cfg['in_channel']
        in_channels_list = [
            in_channels_stage2 * 2,
            in_channels_stage2 * 4,
            in_channels_stage2 * 8,
        ]
        out_channels = cfg['out_channel']
        self.max_size = cfg['max_size']
        self.target_size = cfg['target_size']
        self.fpn = RetinaFaceFPN(in_channels_list,out_channels)
        self.ssh1 = SSH(out_channels, out_channels)
        self.ssh2 = SSH(out_channels, out_channels)
        self.ssh3 = SSH(out_channels, out_channels)

        self.class_head = self._make_class_head(fpn_num=3, inchannels=cfg['out_channel'])
        self.bbox_head = self._make_bbox_head(fpn_num=3, inchannels=cfg['out_channel'])
        self.landmark_head = self._make_landmark_head(fpn_num=3, inchannels=cfg['out_channel'])

    def _make_class_head(self,fpn_num=3,inchannels=64,anchor_num=2):
        classhead = nn.ModuleList()
        for _ in range(fpn_num):
            classhead.append(ClassHead(inchannels,anchor_num))
        return classhead

    def _make_bbox_head(self,fpn_num=3,inchannels=64,anchor_num=2):
        bboxhead = nn.ModuleList()
        for _ in range(fpn_num):
            bboxhead.append(BboxHead(inchannels,anchor_num))
        return bboxhead

    def _make_landmark_head(self,fpn_num: int=3,inchannels: int=64,anchor_num :int=2):
        landmarkhead = nn.ModuleList()
        for _ in range(fpn_num):
            landmarkhead.append(LandmarkHead(inchannels,anchor_num))
        return landmarkhead

    def forward(self,inputs: torch.FloatTensor):

        out = self.body(inputs)

        # FPN
        fpn = self.fpn(out)

        # SSH
        feature1 = self.ssh1(fpn[0])
        feature2 = self.ssh2(fpn[1])
        feature3 = self.ssh3(fpn[2])
        features = [feature1, feature2, feature3]

        bbox_regressions = torch.cat([self.bbox_head[i](feature) for i, feature in enumerate(features)], dim=1)
        classifications = torch.cat([self.class_head[i](feature) for i, feature in enumerate(features)],dim=1)
        ldm_regressions = torch.cat([self.landmark_head[i](feature) for i, feature in enumerate(features)], dim=1)

        if self.phase == 'train':
            output = (bbox_regressions, classifications, ldm_regressions)
        else:
            output = (bbox_regressions, F.softmax(classifications, dim=-1), ldm_regressions)
        return output

    def inference(self, img: np.array) -> Tuple[torch.FloatTensor, torch.FloatTensor]:
        im_shape = img.shape
        h, w, _ = im_shape
        im_height, im_width = h, w

        # pylint: disable=unsubscriptable-object
        scale = torch.Tensor([img.shape[1], img.shape[0], img.shape[1], img.shape[0]])
        device = img.get_device()
        if device == -1:
            device = torch.device('cpu')
        mean = torch.FloatTensor(self.cfg['mean']).to(device)
        img = img - mean
        img = img.permute(2, 0, 1)
        img = img.unsqueeze(0)
        img = img.to(device)

        scale = scale.to(device)

        loc, conf, landms = self.forward(img)
        priorbox = PriorBox(self.cfg, image_size=(im_height, im_width))
        priors = priorbox.forward()
        priors = priors.to(device)
        prior_data = priors.data
        boxes = decode(loc.data.squeeze(0), prior_data, self.cfg['variance'])
        boxes = boxes * scale
        boxes = boxes.to(device)
        scores = conf.squeeze(0).data[:, 1]
        landms = decode_landm(landms.data.squeeze(0), prior_data, self.cfg['variance'])

        scale_replicate = torch.Tensor([img.shape[3], img.shape[2]] * 5)
        scale_replicate = scale_replicate.to(device)
        dets = torch.hstack((boxes, scores.unsqueeze(-1)))
        keep = torchvision.ops.nms(dets[:,:4], dets[:,4], self.cfg['nms_threshold'])
        dets = dets[keep, :]
        landms = landms[keep]
        confident_ids = torch.where(dets[:,4] > self.cfg['confidence_threshold'])
        return dets[confident_ids], landms[confident_ids]





# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.dag.graph_repr import GraphRepr
from towhee.dataframe import DataFrame
from towhee.errors import NoSchedulerError, EmptyInputError
from towhee.engine.graph_context import GraphContext
from towhee.engine.task_scheduler import TaskScheduler


class Pipeline:
    """
    The runtime pipeline context, include graph context, all dataframes.

    Args:
        graph_repr: (`str` or `towhee.dag.GraphRepr`)
            The graph representation either as a YAML-formatted string, or directly
            as an instance of `GraphRepr`.
        parallelism: (`int`)
            The parallelism parameter dictates how many copies of the graph context
            we create. This is likely a low number (1-4) for local engines, but may
            be much higher for cloud instances.
    """

    def __init__(self, graph_repr: GraphRepr, parallelism: int = 1) -> None:
        self._parallelism = parallelism
        self.on_graph_finish_handlers = []
        self._scheduler = None
        self._graph_count = 0

        if isinstance(graph_repr, str):
            self._graph_repr = GraphRepr.from_yaml(graph_repr)
        else:
            self._graph_repr = graph_repr

    @property
    def pipeline_type(self) -> int:
        return self._graph_repr.graph_type

    @property
    def parallelism(self) -> int:
        return self._parallelism

    @property
    def graph_repr(self) -> GraphRepr:
        return self._graph_repr

    @parallelism.setter
    def parallelism(self, val):
        if isinstance(val, int) and val > 0:
            self._parallelism = val
        else:
            raise ValueError('Parallelism value must be a positive integer')

    def register(self, scheduler: TaskScheduler):
        self._scheduler = scheduler

    def __repr__(self) -> str:
        if self._graph_repr._ir is not None:
            return self._graph_repr._ir
        else:
            return object.__repr__(self)

    def __call__(self, inputs: DataFrame) -> DataFrame:
        """
        Process an input `DataFrame`. This function instantiates an output `DataFrame`;
        upon completion, individual `GraphContext` outputs are merged into this
        dataframe. Inputs are weaved through the input `DataFrame` for each
        `GraphContext` as follows (`parallelism = 3`):
            data[0] -> ctx[0]
            data[1] -> ctx[1]
            data[2] -> ctx[2]
            data[3] -> ctx[0]
            data[4] -> ctx[1]

        Args:
            inputs (`towhee.dataframe.DataFrame`):
                Input `DataFrame` (with potentially multiple rows) to process.

        Returns:
            (`towhee.dataframe.DataFrame`)
                Output `DataFrame` with ordering matching the input `DataFrame`.
        """
        if not self._scheduler:
            raise NoSchedulerError('Pipeline not registered to a Scheduler.')

        assert self._parallelism == 1
        graph_ctx = GraphContext(self._graph_count, self._graph_repr)
        self._graph_count += 1
        self._scheduler.register(graph_ctx)
        input_data = inputs.get(0, 1)
        if input_data is None:
            raise EmptyInputError('Input data is empty')
        graph_ctx(input_data[0])
        graph_ctx.join()
        return graph_ctx.result()



# original code from https://github.com/SwinTransformer/Video-Swin-Transformer
# modified by Zilliz.

import torch
from functools import lru_cache

from towhee.models.utils.window_partition import window_partition


# cache each stage results
@lru_cache()
def compute_mask(d, h, w, window_size, shift_size, device):
    """
    Compute attention mask.

    Args:
        d (`int`):
            lenth in time space.
        h (`int`):
            height of image.
        w (`int`):
            width of image.
        window_size (`tuple[int]`):
            window size of W-MSA.
        shift_size (`tuple[int]`):
            shift size of window.
        device (`int`):
            device id.

    Returns:
        attn_mask (`torch.Tensor`):
            Computed attention mask.
    """
    img_mask = torch.zeros((1, d, h, w, 1), device=device)  # 1 Dp Hp Wp 1
    cnt = 0
    for di in slice(-window_size[0]), slice(-window_size[0], -shift_size[0]), slice(-shift_size[0], None):
        for hi in slice(-window_size[1]), slice(-window_size[1], -shift_size[1]), slice(-shift_size[1], None):
            for wi in slice(-window_size[2]), slice(-window_size[2], -shift_size[2]), slice(-shift_size[2], None):
                img_mask[:, di, hi, wi, :] = cnt
                cnt += 1
    mask_windows = window_partition(img_mask, window_size)  # nW, ws[0]*ws[1]*ws[2], 1
    mask_windows = mask_windows.squeeze(-1)  # nW, ws[0]*ws[1]*ws[2]
    attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
    attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(attn_mask == 0, float(0.0))
    return attn_mask



# original code from https://github.com/SwinTransformer/Video-Swin-Transformer
# modified by Zilliz.

from torch import nn
import torch
from towhee.models.utils.weight_init import trunc_normal_


class WindowAttention3D(nn.Module):
    """
    Window based multi-head self attention (W-MSA) module with relative position bias.
    It supports both of shifted and non-shifted window.
    ref1: https://arxiv.org/pdf/2010.11929.pdf
    ref2: https://arxiv.org/pdf/2106.13230.pdf
    Args:
        dim (`int`):
            Number of input channels, see ref2.
        window_size (`tuple[int]`):
            The temporal length, height and width of the window, see ref2.
        num_heads (`int`):
            Number of attention heads of MSA, see ref1.
        qkv_bias (`bool`):
            If True, add a learnable bias to query, key, value. Default: True
        qk_scale (`float`):
            Override default qk scale of head_dim ** -0.5 if set
        attn_drop (`float`):
            Dropout ratio of attention weight. Default: 0.0
        proj_drop (`float`):
            Dropout ratio of output. Default: 0.0
    """

    def __init__(self, dim, window_size, num_heads, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):

        super().__init__()
        self.dim = dim
        self.window_size = window_size  # Wd, Wh, Ww
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim ** -0.5

        # define a parameter table of relative position bias
        # 2*Wd-1 * 2*Wh-1 * 2*Ww-1, nH
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1) * (2 * window_size[2] - 1), num_heads))

        # get pair-wise relative position index for each token inside the window
        coords_d = torch.arange(self.window_size[0])
        coords_h = torch.arange(self.window_size[1])
        coords_w = torch.arange(self.window_size[2])
        coords = torch.stack(torch.meshgrid(coords_d, coords_h, coords_w))  # 3, Wd, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 3, Wd*Wh*Ww
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 3, Wd*Wh*Ww, Wd*Wh*Ww
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wd*Wh*Ww, Wd*Wh*Ww, 3
        relative_coords[:, :, 0] += self.window_size[0] - 1  # shift to start from 0
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 2] += self.window_size[2] - 1

        relative_coords[:, :, 0] *= (2 * self.window_size[1] - 1) * (2 * self.window_size[2] - 1)
        relative_coords[:, :, 1] *= (2 * self.window_size[2] - 1)
        relative_position_index = relative_coords.sum(-1)  # Wd*Wh*Ww, Wd*Wh*Ww
        self.register_buffer("relative_position_index", relative_position_index)

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        trunc_normal_(self.relative_position_bias_table, std=.02)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, mask=None):
        """
        Forward function of window attention 3d.
        Args:
            x (`tensor`):
                Input features with shape of (num_windows*b, n, c). See ref2, n = (t * h * w) / (p * m * m).
                num_windows is Wd * Wh * Ww.
                b is batch size. c is dimension or channels size, same as dim param in init function.
            mask (`tensor`):
                Mask with shape of (num_windows, n, n) or None
        """
        b, n, c = x.shape
        qkv = self.qkv(x).reshape(b, n, 3, self.num_heads, c // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]  # B_, nH, N, C

        q = q * self.scale
        attn = q @ k.transpose(-2, -1)

        relative_position_bias = self.relative_position_bias_table[self.relative_position_index[:n, :n].reshape(-1)]\
            .reshape(
            n, n, -1)  # Wd*Wh*Ww,Wd*Wh*Ww,nH
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wd*Wh*Ww, Wd*Wh*Ww
        attn = attn + relative_position_bias.unsqueeze(0) # B_, nH, N, N

        if mask is not None:
            nw = mask.shape[0]
            attn = attn.view(b // nw, nw, self.num_heads, n, n) + mask.unsqueeze(1).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, n, n)
            attn = self.softmax(attn)
        else:
            attn = self.softmax(attn)

        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(b, n, c)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x



# Original pytorch implementation by:
# 'An Image Is Worth 16 x 16 Words: Transformers for Image Recognition at Scale'
#       - https://arxiv.org/abs/2010.11929
# 'How to train your ViT? Data, Augmentation, and Regularization in Vision Transformers'
#       - https://arxiv.org/abs/2106.10270
#
# Original code by / Copyright 2020, Ross Wightman.
# Modifications & additions by / Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
from torch import nn
from functools import partial
from collections import OrderedDict

from towhee.models.utils.init_vit_weights import init_vit_weights
from towhee.models.layers.vit_block import Block
from towhee.models.layers.patch_embed2d import PatchEmbed2D
from towhee.models.vit.vit_utils import get_configs, load_pretrained_model


class VitModel(nn.Module):
    """
    Vision Transformer Model
    Args:
        name (str): model name
        weights_path (str): path to model weights
        pretrained (bool): if model is pretrained
        img_size (int): image height or width (height=width)
        patch_size (int): patch height or width (height=width)
        in_c (int): number of image channels
        num_classes (int): number of classes
        embed_dim (int): number of features
        depth (int): number of blocks
        num_heads (int): number of heads for Multi-Attention layer
        mlp_ratio (int): mlp ratio
        qkv_bias (bool): if add bias to qkv layer
        qk_scale (float): number to scale qk
        representation_size (int): size of representations
        drop_ratio (float): drop rate of a block
        attn_drop_ratio (float): drop rate of attention layer
        drop_path_ratio (float): drop rate of drop_path layer
        embed_layer: patch embedding layer
        norm_layer: normalization layer
        act_layer: activation layer
    """
    def __init__(self,
                 name=None, weights_path=None, pretrained=False,
                 img_size=224, patch_size=16, in_c=3, num_classes=1000,
                 embed_dim=768, depth=12, num_heads=12, mlp_ratio=4, qkv_bias=True, qk_scale=None,
                 representation_size=None, drop_ratio=0, attn_drop_ratio=0, drop_path_ratio=0,
                 embed_layer=PatchEmbed2D, norm_layer=None, act_layer=None):
        super().__init__()
        self.num_classes = num_classes
        self.num_features = self.embed_dim = embed_dim  # num_features for consistency with other models
        self.num_tokens = 1
        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)
        act_layer = act_layer or nn.GELU

        self.patch_embed = embed_layer(img_size=img_size, patch_size=patch_size, in_chans=in_c, embed_dim=embed_dim)
        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + self.num_tokens, embed_dim))
        self.pos_drop = nn.Dropout(p=drop_ratio)

        dpr = [x.item() for x in torch.linspace(0, drop_path_ratio, depth)]  # stochastic depth decay rule
        self.blocks = nn.Sequential(*[
            Block(dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias, qk_scale=qk_scale,
                  drop_ratio=drop_ratio, attn_drop_ratio=attn_drop_ratio, drop_path_ratio=dpr[i],
                  norm_layer=norm_layer, act_layer=act_layer)
            for i in range(depth)
        ])
        self.norm = norm_layer(embed_dim)

        # Representation layer
        if representation_size:
            self.has_logits = True
            self.num_features = representation_size
            self.pre_logits = nn.Sequential(OrderedDict([
                ("fc", nn.Linear(embed_dim, representation_size)),
                ("act", nn.Tanh())
            ]))
        else:
            self.has_logits = False
            self.pre_logits = nn.Identity()

        # Classifier head(s)
        self.head = nn.Linear(self.num_features, num_classes) if num_classes > 0 else nn.Identity()

        # Weight init
        nn.init.trunc_normal_(self.pos_embed, std=0.02)
        nn.init.trunc_normal_(self.cls_token, std=0.02)
        self.apply(init_vit_weights)

        if pretrained:
            configs = get_configs(name)
            load_pretrained_model(self, configs, weights_path=weights_path)

    def forward_features(self, x):
        # [B, C, H, W] -> [B, num_patches, embed_dim]
        x = self.patch_embed(x)  # [B, 196, 768]
        # [1, 1, 768] -> [B, 1, 768]
        cls_token = self.cls_token.expand(x.shape[0], -1, -1)
        x = torch.cat((cls_token, x), dim=1)  # [B, 197, 768]

        x = self.pos_drop(x + self.pos_embed)
        x = self.blocks(x)
        x = self.norm(x)
        return self.pre_logits(x[:, 0])

    def forward(self, x):
        x = self.forward_features(x)
        x = self.head(x)
        return x



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


from torch import nn
from towhee.models.utils.weight_init import trunc_normal_, lecun_normal_


def window_partition(x, window_size: int):
    """
    Args:
       x: (b, h, w, c)
       window_size (int): window size
    Returns:
       windows: (num_windows*B, window_size, window_size, c)
    """
    b, h, w, c = x.shape
    x = x.view(b, h // window_size, window_size, w // window_size, window_size, c)
    windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, c)
    return windows


def window_reverse(windows, window_size: int, h: int, w: int):
    """
    Args:
       windows: (num_windows*b, window_size, window_size, c)
       window_size (int): Window size
       h (int): Height of image
       w (int): Width of image
    Returns:
       x: (b, h, w, c)
    """
    b = int(windows.shape[0] / (h * w / window_size / window_size))
    x = windows.view(b, h // window_size, w // window_size, window_size, window_size, -1)
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(b, h, w, -1)
    return x

def init_vit_weights(module: nn.Module, name: str = '', head_bias: float = 0., jax_impl: bool = False):
    """ ViT weight initialization
    * When called without n, head_bias, jax_impl args it will behave exactly the same
      as my original init for compatibility with prev hparam / downstream use cases (ie DeiT).
    * When called w/ valid n (module name) and jax_impl=True, will (hopefully) match JAX impl
    """
    if isinstance(module, nn.Linear):
        if name.startswith('head'):
            nn.init.zeros_(module.weight)
            nn.init.constant_(module.bias, head_bias)
        elif name.startswith('pre_logits'):
            lecun_normal_(module.weight)
            nn.init.zeros_(module.bias)
        else:
            if jax_impl:
                nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    if 'mlp' in name:
                        nn.init.normal_(module.bias, std=1e-6)
                    else:
                        nn.init.zeros_(module.bias)
            else:
                trunc_normal_(module.weight, std=.02)
                if module.bias is not None:
                    nn.init.zeros_(module.bias)
    elif jax_impl and isinstance(module, nn.Conv2d):
        # NOTE conv was left to pytorch default in my original init
        lecun_normal_(module.weight)
        if module.bias is not None:
            nn.init.zeros_(module.bias)
    elif isinstance(module, (nn.LayerNorm, nn.GroupNorm, nn.BatchNorm2d)):
        nn.init.zeros_(module.bias)
        nn.init.ones_(module.weight)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from dataclasses import asdict, dataclass, field, fields
from enum import Enum

import torch
from typing import Optional, Dict, Any, Union

from towhee.utils.log import trainer_log
from towhee.utils.yaml_utils import dump_yaml, load_yaml

HELP = "help"
CATEGORY = "category"
OTHER_CATEGORY = "other"


def _get_attr_str(obj, attr_name):
    fld = getattr(obj, attr_name)
    if isinstance(fld, Enum):
        return fld.value
    else:
        return fld


def dump_default_yaml(yaml_path):
    """
    Dump a default yaml, which can be overridden by the custom operator.
    """
    training_config = TrainingConfig()
    training_config.save_to_yaml(path2yaml=yaml_path)
    trainer_log.info("dump default yaml to %s", yaml_path)


def get_config_help():
    """
    Get config setting infos.
    Returns:
        (`dict`)
            The help dict.
    """
    config_fields = fields(TrainingConfig)
    help_dict = {}
    for config_field in config_fields:
        metadata_dict = config_field.metadata
        help_dict[config_field.name] = metadata_dict
    for field_name, metadata_dict in help_dict.items():
        print("- " + field_name)
        print(metadata_dict)
        print("")
    return help_dict


def _early_stopping_factory():
    return {"monitor": "eval_epoch_metric", "patience": 4, "mode": "max"}


def _tensorboard_factory():
    return {
        "log_dir": None,  # if None, the tensorboard will make a dir `./runs` as the log_dir
        "comment": ""
    }


def _model_checkpoint_factory():
    return {"every_n_epoch": 1}


@dataclass
class TrainingConfig:
    """
    The training config, it can be defined in a yaml file

    Args:
        output_dir (`str`):
            The output directory where the model predictions and checkpoints will be written.
        overwrite_output_dir (`bool`):
            Overwrite the content of the output directory.
        eval_strategy (`str`):
            The evaluation strategy.
        eval_steps (`int`):
            Run an evaluation every X steps.
        batch_size (`int`):
            Batch size for training.
        val_batch_size (`int`):
            Batch size for evaluation.
        seed (`int`):
            Random seed that will be set at the beginning of training.
        epoch_num (`int`):
            Total number of training epochs to perform.
        dataloader_pin_memory (`bool`):
            Drop the last incomplete batch if it is not divisible by the batch size.
        dataloader_drop_last (`bool`):
            Drop the last incomplete batch if it is not divisible by the batch size.
        dataloader_num_workers (`int`):
            Number of subprocesses to use for data loading.
        lr (`float`):
            The initial learning rate for AdamW.
        metric (`str`):
            The metric to use to compare two different models.
        print_steps (`int`):
            If None, use the tqdm progress bar, otherwise it will print the logs on the screen every `print_steps`.
        load_best_model_at_end (`bool`):
            Whether or not to load the best model found during training at the end of training.
        early_stopping (`Union[dict, str]`):
            Early stopping.
        model_checkpoint (`Union[dict, str]`):
            Model checkpoint.
        tensorboard (`Union[dict, str]`):
            Tensorboard.
        loss (`Union[str, Dict[str, Any]]`):
            Pytorch loss in torch.nn package.
        optimizer (`Union[str, Dict[str, Any]]`):
            Pytorch optimizer Class name in torch.optim package.
        lr_scheduler_type (`str`):
            The scheduler type to use.
        warmup_ratio (`float`):
            Linear warmup over warmup_ratio fraction of total steps.
        device_str (`str`):
            Device string.
        sync_bn (`bool`):
            It will be work if device_str is `cuda`, the True sync_bn would make training slower but acc better.
        freeze_bn (`bool`):
            It will completely freeze all BatchNorm layers during training.
    """
    output_dir: str = field(
        default="./output_dir",
        metadata={
            HELP: "The output directory where the model predictions and checkpoints will be written.", CATEGORY: "train"
        },
    )
    overwrite_output_dir: bool = field(
        default=True,
        metadata={
            HELP: ("Overwrite the content of the output directory."
                   "Use this to continue training if output_dir points to a checkpoint directory."),
            CATEGORY: "train"
        },
    )
    eval_strategy: str = field(
        default="epoch",
        metadata={
            HELP: "The evaluation strategy. It can be `steps`, `epoch`, `eval_epoch` or `no`,", CATEGORY: "train"
        },
    )
    eval_steps: int = field(default=None, metadata={HELP: "Run an evaluation every X steps.", CATEGORY: "train"})
    batch_size: Optional[int] = field(default=8, metadata={HELP: "Batch size for training.", CATEGORY: "train"})
    val_batch_size: Optional[int] = field(default=-1, metadata={HELP: "Batch size for evaluation.", CATEGORY: "train"})
    seed: int = field(default=42,
                      metadata={HELP: "Random seed that will be set at the beginning of training.", CATEGORY: "train"})

    epoch_num: int = field(default=2, metadata={HELP: "Total number of training epochs to perform.", CATEGORY: "train"})
    dataloader_pin_memory: bool = field(default=True, metadata={HELP: "Whether or not to pin memory for DataLoader.",
                                                                CATEGORY: "train"})
    dataloader_drop_last: bool = field(
        default=True, metadata={
            HELP: "Drop the last incomplete batch if it is not divisible by the batch size.", CATEGORY: "train"
        }
    )
    dataloader_num_workers: int = field(
        default=0,
        metadata={
            HELP:
                "Number of subprocesses to use for data loading."
                "default 0 means that the data will be loaded in the main process."
                "-1 means using all the cpu kernels,"
                "it will greatly improve the speed when distributed training.",
            CATEGORY:
                "train"
        },
    )
    lr: float = field(default=5e-5, metadata={HELP: "The initial learning rate for AdamW.", CATEGORY: "learning"})
    metric: Optional[str] = field(default="Accuracy",
                                  metadata={HELP: "The metric to use to compare two different models.",
                                            CATEGORY: "metrics"})

    print_steps: Optional[int] = field(
        default=None,
        metadata={
            HELP: "if None, use the tqdm progress bar, otherwise it will print the logs on the screen every `print_steps`",
            CATEGORY: "logging"
        }
    )
    load_best_model_at_end: Optional[bool] = field(
        default=False,
        metadata={
            HELP: "Whether or not to load the best model found during training at the end of training.",
            CATEGORY: "train"
        },
    )
    early_stopping: Union[dict, str] = field(default_factory=_early_stopping_factory, metadata={
        HELP: "If the metrics is not better than before in several epoch, it will stop the training.",
        CATEGORY: "callback"}
                                             )
    model_checkpoint: Union[dict, str] = field(default_factory=_model_checkpoint_factory,
                                               metadata={HELP: "How many n epoch to save checkpoints",
                                                         CATEGORY: "callback"})
    tensorboard: Optional[Union[dict, str]] = field(default_factory=_tensorboard_factory,
                                                    metadata={HELP: "Tensorboard.", CATEGORY: "callback"})
    loss: Union[str, Dict[str, Any]] = field(default="CrossEntropyLoss",
                                             metadata={HELP: "Pytorch loss in torch.nn package", CATEGORY: "learning"})
    optimizer: Union[str, Dict[str, Any]] = field(
        default="Adam", metadata={
            HELP: "Pytorch optimizer Class name in torch.optim package", CATEGORY: "learning"
        }
    )
    lr_scheduler_type: str = field(
        default="linear",
        metadata={
            HELP: ("The scheduler type to use."
                   "eg. `linear`, `cosine`, `cosine_with_restarts`, `polynomial`, `constant`, `constant_with_warmup`"),
            CATEGORY: "learning"
        },
    )
    warmup_ratio: float = field(default=0.0, metadata={HELP: "Linear warmup over warmup_ratio fraction of total steps.",
                                                       CATEGORY: "learning"})
    warmup_steps: int = field(default=0, metadata={HELP: "Linear warmup over warmup_steps.", CATEGORY: "learning"})
    device_str: str = field(
        default=None,
        metadata={
            HELP:
                ("None -> if there is a cuda env in the machine, it will use cuda:0, else cpu;"
                 "`cpu` -> use cpu only;"
                 "`cuda:2` -> use the No.2 gpu."),
            CATEGORY:
                "device"
        }
    )
    sync_bn: bool = field(default=False, metadata={
        HELP: "will be work if device_str is `cuda`, the True sync_bn would make training slower but acc better.",
        CATEGORY: "device"
    })
    freeze_bn: bool = field(default=False,
                            metadata={HELP: "will completely freeze all BatchNorm layers during training.",
                                      CATEGORY: "train"})

    def __post_init__(self):
        if self.output_dir is not None:
            self.output_dir = os.path.expanduser(self.output_dir)
        self.should_save = True
        self._get_config_categories()

    @property
    def train_batch_size(self) -> int:
        assert self.batch_size > 0
        train_batch_size = self.batch_size  # * max(1, self.n_gpu)
        return train_batch_size

    @property
    def eval_batch_size(self) -> int:
        assert self.batch_size > 0
        if self.val_batch_size == -1:
            return self.batch_size  # * max(1, self.n_gpu)
        else:
            return self.val_batch_size

    def to_dict(self):
        return asdict(self)

    def to_json_string(self):
        return json.dumps(self.to_dict(), indent=2)

    @property
    def device(self) -> "torch.device":
        if self.device_str is None:
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        else:
            device = torch.device(self.device_str)
        return device

    def load_from_yaml(self, path2yaml: str = None) -> "TrainingConfig":
        """
        Load training configuration from yaml.

        Args:
            path2yaml (`str`):
                The path to yaml.
        Return:
            (`TrainingConfig`).
                TrainingConfig instance self.

        Example:
            >>> from towhee.trainer.training_config import TrainingConfig
            >>> from pathlib import Path
            >>> conf = Path(__file__).parent / "config.yaml"
            >>> ta = TrainingConfig()
            >>> ta.save_to_yaml(conf)
            >>> ta.load_from_yaml(conf)
            >>> ta.epoch_num
            2
        """
        with open(path2yaml, "r", encoding="utf-8") as f:
            config_dict = load_yaml(f)
            for file_category in config_dict:
                if file_category not in self.config_category_set:
                    trainer_log.warning("category %s is not a attribute in TrainingConfig", file_category)
            for category in self.config_category_set:
                category_dict = config_dict.get(category, {})
                self._set_attr_from_dict(category_dict)
            return self

    def save_to_yaml(self, path2yaml: str = None):
        """
        Save training configuration to yaml.

        Args:
            path2yaml (`str`):
                The path to yaml.

        Example:
            >>> from towhee.trainer.training_config import TrainingConfig
            >>> from pathlib import Path
            >>> conf = Path(__file__).parent / 'config.yaml'
            >>> ta = TrainingConfig()
            >>> ta.save_to_yaml(conf)
            >>> ta.load_from_yaml(conf)
            >>> ta.epoch_num
            2
        """
        config_dict = {}
        for config_category in self.config_category_set:
            config_dict[config_category] = {}
        config_fields = fields(TrainingConfig)
        for config_field in config_fields:
            metadata_dict = config_field.metadata
            field_name = config_field.name
            if CATEGORY in metadata_dict:
                category = metadata_dict[CATEGORY]
                config_dict[category][field_name] = _get_attr_str(self, field_name)
            else:
                config_dict[OTHER_CATEGORY][field_name] = _get_attr_str(self, field_name)
                trainer_log.warning("metadata in self.%s has no CATEGORY", config_field.name)
        with open(path2yaml, "w", encoding="utf-8") as file:
            dump_yaml(config_dict, file)

    def _set_attr_from_dict(self, train_config_dict):
        for key, value in train_config_dict.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                trainer_log.warning("TrainingConfig has no attribute %s", key)

    def _get_config_categories(self):
        self.config_category_set = set()
        config_fields = fields(TrainingConfig)
        for config_field in config_fields:
            metadata_dict = config_field.metadata
            if CATEGORY in metadata_dict:
                self.config_category_set.add(metadata_dict[CATEGORY])
            else:
                self.config_category_set.add(OTHER_CATEGORY)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.engine import register

# pylint: disable=import-outside-toplevel
# pylint: disable=invalid-name


@register(name='builtin/image_load')
def image_load(file_or_buff):
    """load image from paths, file objects or memory blocks.

    Returns:
        ndarray: output image

    Examples:

    >>> from towhee.hub import preclude
    >>> from towhee.functional import DataCollection
    >>> dc = (
    ...     DataCollection.range(5)
    ...         .tensor_random(shape=[100, 100, 3])
    ...         .image_dump()
    ... ).to_list()

    >>> (
    ...     DataCollection(dc).image_load()
    ...         .map(lambda x: x.shape)
    ... ).to_list()
    [(100, 100, 3), (100, 100, 3), (100, 100, 3), (100, 100, 3), (100, 100, 3)]
    """
    from towhee.utils.cv2_utils import cv2
    import numpy as np

    if isinstance(file_or_buff, str):
        return cv2.imread(file_or_buff)
    if hasattr(file_or_buff, 'read'):
        buff = np.frombuffer(file_or_buff.read(), dtype=np.uint8)
    else:
        buff = np.frombuffer(file_or_buff, dtype=np.uint8)
    return cv2.imdecode(buff, 1)


@register(name='builtin/image_dump')
class image_dump:
    """dump image to a binary buffer.
    """

    def __init__(self, ext='.JPEG'):
        self._ext = ext

    def __call__(self, x):
        from towhee.utils.cv2_utils import cv2

        retval, im = cv2.imencode(self._ext, x)
        if retval >= 0:
            return im
        else:
            raise Exception('error while encoding image: {}'.format(retval))


@register(name='builtin/image_resize')
class image_resize:
    """resize an image.

    Args:
        dsize ((int, int), optional): target image size. Defaults to None.
        fx (float, optional): scale factor for x axis. Defaults to None.
        fy (float, optional): scale factor for y axis. Defaults to None.
        interpolation (str|int, optional): interpolation method, see detailed document for `cv2.resize`. Defaults to None.

    Returns:
        ndarray: output image.

    Examples:

    >>> from towhee.functional import DataCollection
    >>> dc = (
    ...     DataCollection.range(5)
    ...         .tensor_random(shape=[100, 100, 3])
    ...         .image_resize(dsize=[10, 10], interpolation='nearest')
    ... )
    >>> dc.select['shape']().as_raw().to_list()
    [(10, 10, 3), (10, 10, 3), (10, 10, 3), (10, 10, 3), (10, 10, 3)]
    """

    def __init__(self, dsize=None, fx=None, fy=None, interpolation=None):
        from towhee.utils.cv2_utils import cv2

        self._dsize = dsize
        self._fx = fx
        self._fy = fy
        if interpolation is not None and isinstance(interpolation, str):
            interpolation = 'INTER_{}'.format(interpolation).upper()
            if hasattr(cv2, interpolation):
                interpolation = getattr(cv2, interpolation)
            else:
                interpolation = None
        self._interpolation = interpolation

    def __call__(self, x):
        from towhee.utils.cv2_utils import cv2
        return cv2.resize(x,
                          dsize=self._dsize,
                          fx=self._fx,
                          fy=self._fy,
                          interpolation=self._interpolation)


@register(name='builtin/image_convert_color')
class image_convert_color:
    """convert image color space.

    Args:
        code (str|int): color space conversion string or code

    Returns:
        ndarray: output image.

    Examples:

    >>> from towhee.functional import DataCollection
    >>> import numpy as np
    >>> (
    ...     DataCollection([np.ones([1,1, 3], dtype=np.uint8)])
    ...         .image_convert_color(code='rgb2gray')
    ...         .to_list()
    ... )
    [array([[1]], dtype=uint8)]
    """

    def __init__(self, code):
        from towhee.utils.cv2_utils import cv2

        if code is not None and isinstance(code, str):
            code = 'COLOR_{}'.format(code).upper()
            if hasattr(cv2, code):
                code = getattr(cv2, code)
            else:
                code = None
        self._code = code

    def __call__(self, x):
        from towhee.utils.cv2_utils import cv2
        return cv2.cvtColor(x, self._code)


@register(name='builtin/image_filter')
class image_filter:
    """image filter.
    """

    def __init__(self, ddepth, kernel):
        self._ddepth = ddepth
        self._kernel = kernel

    def __call__(self, x):

        from towhee.utils.cv2_utils import cv2

        return cv2.filter2D(x, self._ddepth, self._kernel)


@register(name='builtin/image_blur')
class image_blur:
    """image blur.

    Args:
        ksize ([int]): kernel size, for example: [3, 3].

    Returns:
        ndarray: output image

    >>> from towhee.functional import DataCollection
    >>> import numpy as np
    >>> (
    ...     DataCollection([np.ones([5,5], dtype=np.uint8)])
    ...         .image_blur(ksize=[3,3])
    ...         .to_list()
    ... )
    [array([[1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1],
           [1, 1, 1, 1, 1]], dtype=uint8)]
    """

    def __init__(self, ksize):
        self._ksize = ksize

    def __call__(self, x):
        from towhee.utils.cv2_utils import cv2

        return cv2.blur(x, self._ksize)


__test__ = {'image_load': image_load.__doc__}

if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False)
